# frozen_string_literal: true

require_relative '../lib/polynomial_table'

RSpec.describe SPARQLprov::Polynomial do
  describe '::constant' do
    it 'creates a constant polynomial zero' do
      expect(SPARQLprov::Polynomial.constant(0).to_s).to eq('0')
    end

    it 'creates a constant polynomial one' do
      expect(SPARQLprov::Polynomial.constant(1).to_s).to eq('1')
    end
  end

  describe '::zero' do
    it 'creates a polynomial zero' do
      expect(SPARQLprov::Polynomial.zero.to_s).to eq('0')
    end
  end

  describe '::one' do
    it 'creates a polynomial one' do
      expect(SPARQLprov::Polynomial.one.to_s).to eq('1')
    end
  end

  describe '::variable' do
    it 'creates a polynomial variable' do
      expect(SPARQLprov::Polynomial.variable('x').to_s).to eq('<x>')
    end
  end

  describe '::sum' do
    it 'creates a polynomial sum' do
      p1 = SPARQLprov::Polynomial.one
      p2 = SPARQLprov::Polynomial.variable('x')
      p3 = SPARQLprov::Polynomial.variable('y')
      expect(SPARQLprov::Polynomial.sum([p1, p2]).to_s).to eq('( + 1 <x> )')
      expect(SPARQLprov::Polynomial.sum([p1, p2, p3]).to_s).to eq('( + 1 <x> <y> )')
    end
  end

  describe '::product' do
    it 'creates a polynomial product' do
      p1 = SPARQLprov::Polynomial.one
      p2 = SPARQLprov::Polynomial.variable('x')
      p3 = SPARQLprov::Polynomial.variable('y')
      expect(SPARQLprov::Polynomial.product([p1, p2]).to_s).to eq('( * 1 <x> )')
      expect(SPARQLprov::Polynomial.product([p1, p2, p3]).to_s).to eq('( * 1 <x> <y> )')
    end
  end

  describe '::difference' do
    it 'creates a polynomial product' do
      p1 = SPARQLprov::Polynomial.one
      p2 = SPARQLprov::Polynomial.variable('x')
      expect(SPARQLprov::Polynomial.difference(p1, p2).to_s).to eq('( - 1 <x> )')
    end
  end
end

RSpec.describe SPARQLprov::ProvenanceVariable do
  describe '#root' do
    it 'check if a variable can be root' do
      expect(SPARQLprov::ProvenanceVariable.new('sum').root?).to eq(true)
      expect(SPARQLprov::ProvenanceVariable.new('sum_product').root?).to eq(false)
    end
  end

  describe '#prefix' do
    it 'check if a variable is a prefix' do
      var1 = SPARQLprov::ProvenanceVariable.new('sum')
      var2 = SPARQLprov::ProvenanceVariable.new('sum_product')
      expect(var1.prefix?(var2)).to eq(false)
      expect(var2.prefix?(var1)).to eq(true)
    end
  end

  describe '#subvariable' do
    it 'generates the subvariable of a varibale' do
      var1 = SPARQLprov::ProvenanceVariable.new('sum_product')
      var2 = SPARQLprov::ProvenanceVariable.new('product_1_sum')
      var3 = SPARQLprov::ProvenanceVariable.new('product_1_sum_statement')
      expect(var1.subvariable.name).to eq('product')
      expect(var2.subvariable.name).to eq('sum')
      expect(var3.subvariable.name).to eq('sum_statement')
    end
  end

  describe '#child' do
    it 'check if a varibale is a child' do
      var1 = SPARQLprov::ProvenanceVariable.new('product')
      var2 = SPARQLprov::ProvenanceVariable.new('product_1_sum')
      var3 = SPARQLprov::ProvenanceVariable.new('product_1_sum_statement')
      expect(var1.child?(var2)).to eq(true)
      expect(var1.child?(var3)).to eq(false)
    end
  end
end

RSpec.describe SPARQLprov::PolynomialTable do
  describe '#root' do
    it 'finds the root variable of a table' do
      variable_names = %w[product product_1_statement product_2_sum product_2_sum_statement]
      variables = variable_names.map { |name| SPARQLprov::ProvenanceVariable.new(name) }
      table = SPARQLprov::PolynomialTable.new(variables)
      expect(table.root.name).to eq('product')
    end
  end

  describe '#root_children' do
    it 'finds the root childrens of a table' do
      variable_names = %w[product product_1_statement product_2_sum product_2_sum_statement]
      variables = variable_names.map { |name| SPARQLprov::ProvenanceVariable.new(name) }
      table = SPARQLprov::PolynomialTable.new(variables)
      children = table.root_children.map(&:name).to_set
      expect(children).to eq(%w[product_1_statement product_2_sum].to_set)
    end
  end

  describe '#branch_variables' do
    it 'finds the branch variables of a parent variable' do
      variable_names = %w[product product_1_statement product_2_sum product_2_sum_statement]
      variables = variable_names.map { |name| SPARQLprov::ProvenanceVariable.new(name) }
      table = SPARQLprov::PolynomialTable.new(variables)
      parent1 = SPARQLprov::ProvenanceVariable.new('product')
      branch_variables1 = table.branch_variables(parent1).map(&:name).to_set
      expected_branch_variables1 =
        %w[product product_1_statement product_2_sum product_2_sum_statement].to_set
      expect(branch_variables1).to eq(expected_branch_variables1)
      parent2 = SPARQLprov::ProvenanceVariable.new('product_2_sum')
      branch_variables2 = table.branch_variables(parent2).map(&:name).to_set
      expected_branch_variables2 = %w[product_2_sum product_2_sum_statement].to_set
      expect(branch_variables2).to eq(expected_branch_variables2)
    end
  end

  describe '#value' do
    it 'gets the value in a table cell' do
      variable_names = %w[product product_1_statement product_2_statement]
      variables = variable_names.map { |name| SPARQLprov::ProvenanceVariable.new(name) }
      table = SPARQLprov::PolynomialTable.new(variables)
      value1 = table.value(%w[a b c], variables[0])
      value2 = table.value(%w[a b c], variables[1])
      value3 = table.value(%w[a b c], variables[2])
      expect(value1).to eq('a')
      expect(value2).to eq('b')
      expect(value3).to eq('c')
    end
  end

  describe '#subtables' do
    let(:table) do
      variable_names = %w[product product_1_statement product_2_statement]
      variables = variable_names.map { |name| SPARQLprov::ProvenanceVariable.new(name) }
      table = SPARQLprov::PolynomialTable.new(variables)
      table.add(%w[a b c])
      table.add(%w[a b d])
      table
    end

    it 'gets the subtables for a table variable product_1_statement' do
      variable = SPARQLprov::ProvenanceVariable.new('product_1_statement')
      variables = [SPARQLprov::ProvenanceVariable.new('statement')]
      expected_subtables = {
        'b' => SPARQLprov::PolynomialTable.new(variables)
      }
      expected_subtables['b'].add(%w[b])
      subtables = table.subtables(variable)
      expect(subtables).to eq(expected_subtables)
    end

    it 'gets the subtables for a table variable product_2_statement' do
      variable = SPARQLprov::ProvenanceVariable.new('product_2_statement')
      variables = [SPARQLprov::ProvenanceVariable.new('statement')]
      expected_subtables = {
        'c' => SPARQLprov::PolynomialTable.new(variables),
        'd' => SPARQLprov::PolynomialTable.new(variables)
      }
      expected_subtables['c'].add(%w[c])
      expected_subtables['d'].add(%w[d])
      subtables = table.subtables(variable)
      expect(subtables).to eq(expected_subtables)
    end
  end

  describe '#polynomial' do
    it 'decode a zero polynomial' do
      variable_names = %w[statement]
      variables = variable_names.map { |name| SPARQLprov::ProvenanceVariable.new(name) }
      table = SPARQLprov::PolynomialTable.new(variables)
      expect(table.polynomial.to_s).to eq('0')
      table.add([''])
      expect(table.polynomial.to_s).to eq('0')
    end

    it 'decode a constant polynomial' do
      variable_names = %w[statement]
      variables = variable_names.map { |name| SPARQLprov::ProvenanceVariable.new(name) }
      table = SPARQLprov::PolynomialTable.new(variables)
      table.add(['1'])
      expect(table.polynomial.to_s).to eq('1')
    end

    it 'decode a variable polynomial' do
      variable_names = %w[statement]
      variables = variable_names.map { |name| SPARQLprov::ProvenanceVariable.new(name) }
      table = SPARQLprov::PolynomialTable.new(variables)
      table.add(['x'])
      expect(table.polynomial.to_s).to eq('<x>')
    end

    it 'decode a sum polynomial' do
      variable_names = %w[sum sum_statement]
      variables = variable_names.map { |name| SPARQLprov::ProvenanceVariable.new(name) }
      table = SPARQLprov::PolynomialTable.new(variables)
      table.add(%w[a x])
      table.add(%w[a y])
      expect(table.polynomial.to_s).to eq('( + <x> <y> )')
    end

    it 'decode an osum polynomial' do
      variable_names = %w[osum osum_1_statement osum_2_statement]
      variables = variable_names.map { |name| SPARQLprov::ProvenanceVariable.new(name) }
      table = SPARQLprov::PolynomialTable.new(variables)
      table.add(['a', 'x', ''])
      table.add(['a', 'y', ''])
      table.add(['a', '', 'z'])
      expect(table.polynomial.to_s).to eq('( + ( + <x> <y> 0 ) ( + 0 <z> ) )')
    end

    it 'decode a product polynomial' do
      variable_names = %w[product product_1_statement product_2_statement]
      variables = variable_names.map { |name| SPARQLprov::ProvenanceVariable.new(name) }
      table = SPARQLprov::PolynomialTable.new(variables)
      table.add(%w[a x u])
      table.add(%w[a y u])
      table.add(%w[a x v])
      table.add(%w[a y v])
      expect(table.polynomial.to_s).to eq('( * ( + <x> <y> ) ( + <u> <v> ) )')
    end

    it 'decode a difference polynomial' do
      variable_names = %w[difference difference_1_statement difference_2_sum difference_2_sum_statement]
      variables = variable_names.map { |name| SPARQLprov::ProvenanceVariable.new(name) }
      table = SPARQLprov::PolynomialTable.new(variables)
      table.add(%w[a x b y])
      table.add(%w[a x b z])
      expect(table.polynomial.to_s).to eq('( - <x> ( + <y> <z> ) )')
    end
  end
end
